/******************************************************************************
Program name: drmario
Authors: A. Kaluszka, R. Thomschutz
Date: 2003.12.09
Description:
	Continuously reads in a Dr. Mario game screen, determines playing field
	and pieces within playing field, and signals where to move active piece
Usage:
	drmario
Compiled successfully with MSVC++ 6.0
Libtiff must be installed
Multithreaded C++ run-time library must be enabled
Use linker option: /nodefaultlib:"LIBC"
******************************************************************************/

// dib.h contains Windows functions for screen capture and manipulation
#include "dib.h" 
#include <tiffio.h>
#include <iostream>
#include <cmath>
using namespace std;

#define R 16 // Field is 16 blocks tall
#define C 8 //   and 8 blocks wide
#define R2 3 // The 3 top rows never contain viruses
void printError (char *error);
void flip(int* in_image, int rows, int columns);
void findBox(int* in_image, int rows, int columns, int& x1, int& y1, 
			int& width);
unsigned char grayscale(int input);
char getColor(int* in_image, int x, int y, int columns, int width);
char findVit(char grid1[][C], char grid2[][C], int& x1, int& y1, int& x2, 
			int& y2);
char move(char grid[][C], int x1, int y1, int x2, int y2);

/******************************************************************************
Function name: main
Description: Main loop; manages image conversion to game information
Input parameters: none processed
Returned value: N/A
******************************************************************************/
int main(int argc, char* argv[])
{
	int rows; // number of rows in image
	int columns; // number of columns in image
	int x; // field start x position
	int y; // field start y position
	int width; // field width
	int x1; // vitamin x position (first half)
	int y1; // vitamin y position (first half)
	int x2; // vitamin x position (second half)
	int y2; // vitamin y position (second half)
	char grid[R][C]; // Playing field
	char oldgrid[R][C]; // Previous playing field
	int *in_image; // pointer for input image array
	TIFF *in_filep; // handle for input image file
	// grab top left 640x480 pixels from screen and save as BMP
	SaveDesktopAsFile("x.bmp"); 
	system("convert x.bmp -compress None x.tif"); // convert BMP to TIFF
	in_filep = TIFFOpen("x.tif", "r"); // Open TIFF file
	if(in_filep == NULL) printError ("Could not open input file!");
	// Determine the size of the input image
	TIFFGetField(in_filep, TIFFTAG_IMAGELENGTH, &rows);
	TIFFGetField(in_filep, TIFFTAG_IMAGEWIDTH, &columns);
	// Allocate memory to hold image arrays
	in_image = (int *) _TIFFmalloc(sizeof(int) * (rows * columns));
	if(in_image == NULL) printError ("Could not allocate memory!");
	// Read image into RGB array
	TIFFReadRGBAImage(in_filep, columns, rows, (unsigned long *)in_image, 0);
	TIFFClose(in_filep); // Close file
	// TIFFReadRGBAImage reads from bottom to top, so flip data
	flip(in_image, rows, columns); 
	clrscr(1); // clear screen
	findBox(in_image, rows, columns, x, y, width); // Find playing field
	// Output field location
	cout << "Found field at: " << endl
		<< x << ',' << y << endl 
		<< "with width " << width << ' ' << endl;
	while(1) // Play forever (CTRL+C to quit)
	{
		for(int i = 0; i < R; i++) // Determine pieces for each spot in grid
		{
			for(int j = 0; j < C; j++)
			{
				oldgrid[i][j] = grid[i][j];
				grid[i][j] = getColor(in_image, x + width * j / C, 
					y + width * i / C, columns, width / C);
			}
		}
		// Find active vitamin
		char result = findVit(grid, oldgrid, x1, y1, x2, y2);
		if(result != 0) // Playing field has changed
		{
			clrscr(0); // Refresh screen without flicker
			for(int i = 0; i < R; i++) // Output playing field
			{
				for(int j = 0; j < C; j++) cout << grid[i][j] << ' ';
				cout << endl;
			}
		}
		if(result == 1) // Found vitamin.  Output location and move directions
		{
			cout << "Vitamin at: " << x1 << ',' << y1 << ' ' << x2 << ',' 
				<< y2 << "      " << endl;
			switch(move(grid, x1, y1, x2, y2))
			{
			case 'L':
				cout << "Move left ";
				break;
			case 'R':
				cout << "Move right";
				break;
			default:
				cout << "          ";
			}
		}
		else if(result == -1) // Otherwise, field has changed dramatically
		{
			cout << "Clear detected.                " << endl
				<< "                               " << endl
				<< "                               " << endl
				<< "                               " << endl
				<< "                               " << endl;
		}
		SaveDesktopAsFile("x.bmp"); // Grab screen again
		system("convert x.bmp -compress None x.tif"); // Convert BMP to TIFF
		in_filep = TIFFOpen("x.tif", "r"); // Open TIFF file
		if(in_filep == NULL) printError ("Could not open input file!");
		// Assume size is the same as original image or something is wrong 
		// anyway
		// Read image into RGB array
		TIFFReadRGBAImage(in_filep, columns, rows, 
			(unsigned long *)in_image, 0);
		TIFFClose(in_filep); // Close file
		flip(in_image, rows, columns);
	}
	_TIFFfree(in_image); // Deallocate the image memory and exit
	return 0;
}

/******************************************************************************
Function name: printError
Description: Print an error message, and exit the program.
Input parameters:
	error - pointer to string to be printed
Returned value: none
******************************************************************************/
void printError(char *error)
{
	cerr << "Error:  " << error << endl;
	exit(1);
}

/******************************************************************************
Function name: flip
Description: Flips image top to bottom
Input parameters:
	in_image - input image
	rows - number of rows in image
	columns - number of columns in image
Returned value: in_image becomes flipped
******************************************************************************/
void flip(int* in_image, int rows, int columns)
{
	for(int i = 0; i < rows / 2; i++)
	{
		for(int j = 0; j < columns; j++)
		{
			int temp = in_image[i * columns + j];
			in_image[i * columns + j] = in_image[(rows - i - 1) * columns + j];
			in_image[(rows - i - 1) * columns + j] = temp;
  		}
	}
}

/******************************************************************************
Function name: findBox
Description: Find playing field within image
Input parameters:
	in_image - input image
	rows - number of rows in image
	columns - number of columns in image
	x1/y1/width - passed by reference
Returned value:
	x1 - x coordinate of start of playing field
	y1 - y coordinate of start of playing field
	width - width of playing field
******************************************************************************/
void findBox(int* in_image, int rows, int columns, int& x1, int& y1, 
			 int& width)
{
	x1 = y1 = width = 0;
	for(int r = 0; r < rows / 2; r++) // operate on all pixels
	{
		for(int c = 0; c < columns - width; c++)
		{
			if(r >= y1 && r < y1 + width * R2/C && c >= x1 && c < x1 + width) 
				continue; // skip if within largest found box
			// binary black piece
			if(grayscale(in_image[r * columns + c]) < 128) 
			{
				// if this row of black is shorter than the current max, skip
				int count = 0;
				int c2 = c;
				for(; c2 < columns - width; c2++)
				{
					if(grayscale(in_image[r * columns + c2]) > 127) break;
					count++;
				}
				if(count < width)
				{
					c += count - 1;
					continue;
				}
				// Look for largest 8 x 3 box of dark pixels
				// (The top three rows of the playing field are always clear 
				// at the beginning of the game and the field is 8 blocks wide)
				for(c2 = c; c2 < columns; c2++)
				{
					bool end = 0;
					// only look in top half of image
					for(int r2 = r; r2 - r <= (c2 - c) * R2/C && r2 <= rows; 
						r2++)
					{
						for(int c3 = c; c3 <= c2; c3++)
						{
							// end of box when binary white value is found
							if(grayscale(in_image[r2 * columns + c3]) > 127)
							{
								end = 1;
								break;
							}
       					}
						if(end) break;
					}
					if(end) // If current box is biggest found, save values
					{
						if(c2 - c > width)
						{
							width = c2 - c;
							x1 = c;
							y1 = r;
						}
						break;
					}
				}
			}
		}
	}
}

/******************************************************************************
Function name: grayscale
Description: Find grayscale value of color pixel
Input parameters:
	input - integer representation of color pixel
Returned value: grayscale value of color pixel
******************************************************************************/
unsigned char grayscale(int input)
{
	// Coefficients necessary because human vision does not see each color equally
	return (unsigned char)(0.299 * double(TIFFGetR(input)) + 0.587 * 
		double(TIFFGetG(input)) +	0.114 * double(TIFFGetB(input)));
}

/******************************************************************************
Function name: getColor
Description: Finds active vitamin within playing field or detects major change
Input parameters:
	in_image - input image
	x - x coordinate of start location in image
	y - y coordinate of start location in image
	width - width of piece
Returned value: determined piece (r, y, b, R, Y, B, or .)
	r = red vitamin   y = yellow vitamin  b = blue vitamin
	R = red virus     Y = yellow virus    B = blue virus     . = nothing
******************************************************************************/
char getColor(int* in_image, int x, int y, int columns, int width)
{
	int red = 0; // red average value
	int green = 0; // green average value
	int blue = 0; // blue average value
	int count = 0; // ignored pixel count
	for(int i = y; i < y + width; i++) // look only within specified square
	{
		int red1 = 0; // row red total
		int green1 = 0; // row green total
		int blue1 = 0; // row blue total
		int badcount = 0; // bad pixel count
		for(int j = x; j < x + width; j++)
		{
			// If pixel is not intense enough, it might not be valid
			// This check is necessary since the active vitamin is two pixels
			// lower than the other pieces, so lines must be ignored for
			// calculations to work
			if(grayscale(in_image[i * columns + j]) > 30)
			{
				red1 += TIFFGetR(in_image[i * columns + j]);
				green1 += TIFFGetG(in_image[i * columns + j]);
				blue1 += TIFFGetB(in_image[i * columns + j]);
			}
			else badcount++; // Line is only bad when badcount == width
		}
		if(badcount != width) // Line is valid, add to totals
		{
			red += red1;
			green += green1;
			blue += blue1;
		}
		else count += width; // Add line to ignore count
	}
	// Do not consider if piece does not take up half of the block
	if(width * width - count < width * width / 2) return '.';
	else // normalize values
	{
		red /= (width * width - count);
		green /= (width * width - count);
		blue /= (width * width - count);
	}
	// Determine pieces
	// Red piece: red >> green && red >> blue     (major color is red)
	// Yellow piece: red >> blue && green >> blue (major color is red + green)
	// Blue piece: green >> red && blue >> red    (major color is green + blue)
	// Vitamin: average of major color > 134
	// Virus: 40 < average of major color < 134
	if(red > 1.5 * green && red > 1.5 * blue) // Check for red piece
	{
		if(red > 134) return 'r'; // Higher value is vitamin
		if(red > 40) return 'R'; // Don't falsely detect viruses
	}
	if(red > 1.5 * blue && green > 1.5 * blue) // Check for yellow piece
	{
		// Higher average value is vitamin
		if((red + green) / 2 > 134) return 'y'; 
		if((red + green) / 2 > 40) return 'Y'; // Don't falsely detect viruses
	}
	if(green > 1.5 * red && blue > 1.5 * red) // Check for blue piece
	{
		// Higher average value is vitamin
		if((green + blue) / 2 > 134) return 'b'; 
		if((green + blue) / 2 > 40) return 'B'; // Don't falsely detect viruses
	}
	return '.'; // No piece recognized
}

/******************************************************************************
Function name: findVit
Description: Finds active vitamin within playing field or detects major change
Input parameters:
	grid1 - playing field
	grid2 - previous playing field
	x1/y1/x2/y2 - passed by reference
Returned value:
	-1 if major change has occured, 0 if no change, and 1 for vitamin movement
	x1 - vitamin x position (first half)
	y1 - vitamin y position (first half)
	x2 - vitamin x position (second half)
	y2 - vitamin y position (second half)
******************************************************************************/
char findVit(char grid1[][C], char grid2[][C], int& x1, int& y1, int& x2, 
			 int& y2)
{
	int count = 0; // number of added pieces
	int count1 = 0; // number of removed pieces
	// Look through entire grid for changes
	for(int i = 0; i < R; i++)
	{
		for(int j = 0; j < C; j++)
		{
			if(grid1[i][j] != grid2[i][j])
			{
				if(count == 0)
				{
					x1 = j;
					y1 = i;
				}
				if(count == 1)
				{
					x2 = j;
					y2 = i;
				}
				// if change is not empty space, increase count
				if(grid1[i][j] != '.') count++;
				else count1++;
			}
		}
	}
	if(count < 1) return 0; // No change
	else if(count > 2 || count1 > 2) return -1; // Large change
	else if(count == 1) // Single-color vitamin movement
	{
		// vertical orientation
		if(y1 > 1 
			&& grid1[y1][x1] == grid1[y1 - 1][x1] && grid1[y1 - 2][x1] == '.' 
			&& grid2[y1 - 1][x1] == grid2[y1 - 2][x1] && grid2[y1][x1] == '.')
		{
			x2 = x1;
			y2 = y1--;
		}
		// horizontal moving right
		else if(x1 > 1 
			&& grid1[y1][x1] == grid1[y1][x1 - 1] && grid1[y1][x1 - 2] == '.' 
			&& grid2[y1][x1 - 1] == grid2[y1][x1 - 2] && grid2[y1][x1] == '.')
		{
			x2 = x1--;
			y2 = y1;
		}
		// horizontal moving left
		else if(x1 < C - 1 
			&& grid1[y1][x1] == grid1[y1][x1 + 1] && grid1[y1][x1 + 2] == '.' 
			&& grid2[y1][x1 + 1] == grid2[y1][x1 + 2] && grid2[y1][x1] == '.')
		{
			x2 = x1 + 1;
			y2 = y1;
		}
		// rotation vertical to horizontal
		else if(y1 > 0 && 
			grid1[y1][x1] == grid1[y1][x1 - 1] && grid1[y1 - 1][x1 - 1] == '.' 
			&& grid2[y1][x1 - 1] == grid2[y1 - 1][x1 - 1] 
			&& grid2[y1][x1] == '.')
		{
			x2 = x1--;
			y2 = y1;
		}
		// rotation horizontal to vertical
		else if(y1 > 0 && 
			grid1[y1][x1] == grid1[y1 + 1][x1] && grid1[y1 + 1][x1 + 1] == '.' 
			&& grid2[y1 + 1][x1] == grid2[y1 + 1][x1 + 1] 
			&& grid2[y1][x1] == '.')
		{
			x2 = x1;
			y2 = y1--;
		}
		// rotation out of top of bottle
		else if(y1 == 0 && ((x1 == 3 && grid1[0][4] == '.') 
			|| (x1 == 4 && grid1[0][5] == '.')))
		{
			x2 = x1;
			y2 = y1--;
		}
	}
	return 1;
}

/******************************************************************************
Function name: move
Description: A very simple AI
	This AI does not take into account rotation
	This AI does not look for horizontal matches
	A much better algorithm would find the optimal placement and determine 
	the full path
Input parameters:
	grid - playing field
	x1 - vitamin x position (first half)
	y1 - vitamin y position (first half)
	x2 - vitamin x position (second half)
	y2 - vitamin y position (second half)
Returned value: Direction to move (L = left, R = right, 0 = no movement)
******************************************************************************/
char move(char grid[][C], int x1, int y1, int x2, int y2)
{
	int cmin = 0; // Find limits of movement (walls of objects)
	int cmax = C;
	int max = (y1 > y2 ? y1 : y2);
	// find lefthand boundary
	int i = (x1 < x2 ? x1 : x2);
	for(; i >= 0; i--)
	{
		if(grid[max][i] != '.')
		{
			int limit = i;
			// look for solid wall
			bool nogood = 0;
			for(int j = max + 1; j < R; j++)
			{
				if(grid[j][i] == '.')
				{
					nogood = 1;
					break;
				}
			}
			if(nogood) continue;
			cmin = limit;
			break;
		}
	}
	// find righthand boundary
	i = (x1 > x2 ? x1 : x2);
	for(; i < C; i++)
	{
		if(grid[max][i] != '.')
		{
			int limit = i;
			// look for solid wall
			bool nogood = 0;
			for(int j = max + 1; j < R; j++)
			{
				if(grid[j][i] == '.')
				{
					nogood = 1;
					break;
				}
			}
			if(nogood) continue;
			cmax = limit;
			break;
		}
	}
	// Find placement for vitamin
	// Look for a virus of the same color
	i = cmin;
	for(; i <= cmax; i++)
	{
		for(int j = max + 1; j < R; j++)
		{
			if(grid[j][i] == grid[y2][x2] - 32) // same color, different case
			{
				// Righthand color cannot match to min column without rotation
				// There must not be a piece on top of the left target either
				bool bad = 0; // If something blocks path, go somewhere else
				for(int k = max + 1; k < j; k++)
				{
					if(grid[k][i - 1] != '.') 
					{
						bad = 1;
						break;
					}
				}
				if(i != cmin && !bad)
				{
					cout << "Target: Virus " << i << ',' << j << "       " << endl
						<< "                  " << endl
						<< "Piece: Right side" << endl;
					if(i > x2) return 'R';
					else if(i < x2) return 'L';
					else return '0';
				}
			}
			if(y1 == y2 && grid[j][i] == grid[y1][x1] - 32)
			{   // same color, different case
				// Lefthand color cannot match to max column without rotation
				// There must not be a piece on top of the right target either
				bool bad = 0; // If something blocks path, go somewhere else
				for(int k = max + 1; k < j; k++)
				{
					if(grid[k][i + 1] != '.')
					{
						bad = 1;
						break;
					}
				}
				if(i != cmax && !bad)
				{
					cout << "Target: Virus " << i << ',' << j << "       " << endl
						<< "                     " << endl
						<< "Piece: Left side " << endl;
					if(i > x1) return 'R';
					else if(i < x1) return 'L';
					else return '0';
				}
			}
			if(grid[j][i] != '.') break; // Path is blocked
		}
	}
	// Look for the largest vitamin stack of same color
	i = cmin;
	int maxstack = 0; // max height of same colored stack
	int direction = 0; // holding for i coord of maxstack
	int height = 0; // holding for j coord of maxstack
	int piece = 2; // which half of vitamin
	for(; i <= cmax; i++)
	{
		for(int j = max + 1; j < R; j++)
		{
			if(grid[j][i] == grid[y2][x2])
			{
				// Righthand color cannot match to min column without rotation
				// There must not be a piece on top of the left target either
				bool bad = 0; // If something blocks path, go somewhere else
				int k = max + 1;
				for(; k < j; k++)
				{
					if(grid[k][i - 1] != '.')
					{
						bad = 1;
						break;
					}
				}
				if(i != cmin && !bad)
				{
					int stack = 0;
					for(k = j; k < R; k++) // Find largest stack
					{ // check for both vitamins and viruses of same color
						if(grid[k][i] != grid[y2][x2] 
							&& grid[k][i] != grid[y2][x2] - 32) break;
						stack++;
					}
					if(stack > maxstack)
					{
						maxstack = stack;
						direction = i;
						height = j;
						piece = 2;
					}
				}
			}
			if(y1 == y2 && grid[j][i] == grid[y1][x1])
			{
				// Lefthand color cannot match to max column without rotation
				// There must not be a piece on top of the right target either
				bool bad = 0; // If something blocks path, go somewhere else
				int k = max + 1;
				for(; k < j; k++)
				{
					if(grid[k][i + 1] != '.')
					{
						bad = 1;
						break;
					}
				}
				if(i != cmax && !bad)
				{
					int stack = 0;
					for(k = j; k < R; k++) // Find largest stack
					{ // check for both vitamins and viruses of same color
						if(grid[k][i] != grid[y1][x1] 
							&& grid[k][i] != grid[y1][x1] - 32) break;
						stack++;
					}
					if(stack > maxstack)
					{
						maxstack = stack;
						direction = i;
						height = j;
						piece = 1;
					}
				}
			}
			if(grid[j][i] != '.') break; // Path is blocked
		}
	}
	if(maxstack > 0)
	{
		cout << "Target: Vitamin " << direction << ',' << height << "    " << endl
			<< "Stack: " << maxstack << endl;
		if(piece == 1) // Left part of vitamin
		{
			cout << "Piece: Left side " << endl;
			if(direction > x1) return 'R';
			else if(direction < x1) return 'L';
			else return '0';
		}
		if(piece == 2) // Right part of vitamin
		{
			cout << "Piece: Right side" << endl;
			if(direction > x2) return 'R';
			else if(direction < x2) return 'L';
			else return '0';
		}
	}
	// Find lowest spot in field to place piece if nothing better to do
	int left = 0;
	int bottom = 0;
	i = cmin;
	for(; i <= cmax - 1; i++)
	{
		for(int j = max; j < R; j++)
		{
			if(grid[j][i] != '.' || grid[j][i + 1] != '.')
			{
				if(j > bottom)
				{
					left = i;
					bottom = j;
				}
				break;
			}
		}
	}
	cout << "Target: Nothing " << left << ',' << bottom << "     " << endl
		<< "                " << endl
		<< "                 " << endl;
	if(left > x1) return 'R';
	if(left < x1) return 'L';
	return '0';
}
